package de.java2html.eclipse.popup;

import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;

import de.java2html.converter.IJavaSourceConverter;
import de.java2html.converter.JavaSource2HTMLConverter;
import de.java2html.converter.JavaSourceConverterProvider;
import de.java2html.eclipse.Java2HtmlPlugin;
import de.java2html.eclipse.dialog.Java2HtmlDialog;
import de.java2html.eclipse.util.EclipseUtilities;
import de.java2html.eclipse.util.ReadableDocument;
import de.java2html.eclipse.util.ReadableDocumentCompilationUnitAdapter;
import de.java2html.eclipse.util.ReadableDocumentFileAdapter;
import de.java2html.eclipse.util.ReadableDocumentMethodAdapter;
import de.java2html.javasource.JavaSource;
import de.java2html.javasource.JavaSourceParser;
import de.java2html.options.JavaSourceConversionOptions;
import de.java2html.util.IoUtilities;

import org.eclipse.core.internal.resources.Folder;
import org.eclipse.core.internal.resources.Project;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.core.ClassFile;
import org.eclipse.jdt.internal.core.CompilationUnit;
import org.eclipse.jdt.internal.core.JavaProject;
import org.eclipse.jdt.internal.core.PackageFragment;
import org.eclipse.jdt.internal.core.PackageFragmentRoot;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IActionDelegate;
import org.eclipse.ui.IWorkbenchPart;

/**
 * @author Markus Gebhard
 */
public abstract class AbstractJava2HtmlAction implements IActionDelegate {
  private IWorkbenchPart workbenchPart;

  /**
   * @see org.eclipse.ui.IObjectActionDelegate#setActivePart(org.eclipse.jface.action.IAction, org.eclipse.ui.IWorkbenchPart)
   */
  public void setActivePart(IAction action, IWorkbenchPart part) {
    workbenchPart = part;
  }

  public Shell getShell() {
    return getWorkbenchPart().getSite().getShell();
  }

  protected abstract JavaDocumentList getDocumentList();

  /**
   * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
   */
  public void run(IAction action) {
    JavaDocumentList documents = getDocumentList();
    if (documents == null) {
      return;
    }

    Java2HtmlDialog dialog = new Java2HtmlDialog(getShell());
    int returnCode = dialog.open();
    if (returnCode == Window.CANCEL) {
      return;
    }

    String outputFormat = dialog.getSelectedOutputFormat();
    IJavaSourceConverter converter = JavaSourceConverterProvider.getJavaSourceConverterByName(outputFormat);
    if (dialog.isConvertToFileSelected()) {
      promptForSaveAndConvertDocuments(documents, converter);
    }
    else {
      convertDocumentsToClipboard(documents, converter);
    }
  }

  private void convertDocumentsToClipboard(JavaDocumentList documents, IJavaSourceConverter converter) {
    StringWriter writer = new StringWriter();
    try {
      convertReadableDocuments(documents, converter, writer);
    }
    catch (IOException e) {
      ErrorDialog.openError(
          getShell(),
          "Output error",
          "Error converting selection: " + e.getLocalizedMessage(),
          null);
    }
    String result = writer.getBuffer().toString();

    Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
    clipboard.setContents(new StringSelection(result), new ClipboardOwner() {
      public void lostOwnership(Clipboard clipboard, Transferable contents) {
        //nothing to do
      }
    });
  }

  /**
   * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction, org.eclipse.jface.viewers.ISelection)
   */
  public void selectionChanged(IAction arg0, ISelection arg1) {
    //nothing to do
  }

  protected IWorkbenchPart getWorkbenchPart() {
    return workbenchPart;
  }

  protected JavaDocumentList getJavaDocumentsFromSelection(IStructuredSelection selection) {
    JavaDocumentList list = new JavaDocumentList();

    Iterator iterator = selection.iterator();
    while (iterator.hasNext()) {
      Object obj = iterator.next();
      if (!(obj instanceof IAdaptable)) {
        continue;
      }
      IAdaptable adaptable = (IAdaptable) obj;
      addAllDocumentsToList(list, adaptable);
      if (list.getName() == null) {
        list.setName(EclipseUtilities.getName(adaptable));
      }
    }

    list.removeDuplicates();
    if (list.getName() == null) {
      list.setName("java2html");
    }

    return list;
  }

  protected void addAllDocumentsToList(final JavaDocumentList list, final IAdaptable adaptable) {
    if (adaptable instanceof ClassFile) {
      return;
    }

    if (adaptable instanceof IMethod) {
      list.add(new ReadableDocumentMethodAdapter((IMethod) adaptable));
      return;
    }

    if (adaptable instanceof Folder || adaptable instanceof Project) {
      IResource resource = (IResource) adaptable;
      try {
        resource.accept(new IResourceVisitor() {
          public boolean visit(IResource resource) throws CoreException {
            if (resource != adaptable) {
              addAllDocumentsToList(list, resource);
            }
            return true;
          }
        });
      }
      catch (CoreException e) {
        //nothing to do
      }
      return;
    }

    IResource resource = null;
    File file = null;
    if (adaptable instanceof IResource) {
      resource = (IResource) adaptable;
    }
    else if (adaptable instanceof CompilationUnit) {
      CompilationUnit compilationUnit = (CompilationUnit) adaptable;
      list.add(new ReadableDocumentCompilationUnitAdapter(compilationUnit));
      return;
    }
    else if (adaptable instanceof JavaProject) {
      JavaProject javaProject = (JavaProject) adaptable;
      try {
        IJavaElement[] children = javaProject.getChildren();
        for (int i = 0; i < children.length; ++i) {
          addAllDocumentsToList(list, children[i]);
        }
      }
      catch (JavaModelException e) {
        //nothing to do
      }
      return;
    }
    else if (adaptable instanceof PackageFragment) {
      PackageFragment packageFragment = (PackageFragment) adaptable;
      try {
        IJavaElement[] children = packageFragment.getChildren();
        for (int i = 0; i < children.length; ++i) {
          addAllDocumentsToList(list, children[i]);
        }
      }
      catch (JavaModelException e) {
        //nothing to do
      }
      return;
    }
    else if (adaptable instanceof PackageFragmentRoot) {
      PackageFragmentRoot packageFragmentRoot = (PackageFragmentRoot) adaptable;
      try {
        IJavaElement[] children = packageFragmentRoot.getChildren();
        for (int i = 0; i < children.length; ++i) {
          addAllDocumentsToList(list, children[i]);
        }
      }
      catch (JavaModelException e) {
        //nothing to do
      }
      return;
    }
    else {
      EclipseUtilities.out(getShell(), adaptable.getClass().toString());
    }
    //     else if (adaptable instanceof IDOMCompilationUnit){
    //      resource = ((IDOMCompilationUnit)adaptable).getResource();
    //    } else if (
    //      adaptable instanceof PackageFragment
    //        && ((PackageFragment) adaptable).getPackageFragmentRoot()
    //          instanceof JarPackageFragmentRoot) {
    //      file =
    //        getJarFile(((PackageFragment) adaptable).getPackageFragmentRoot());
    //    } else if (adaptable instanceof JarPackageFragmentRoot) {
    //      file = getJarFile(adaptable);
    //    } else {
    //      resource = (IResource) adaptable.getAdapter(IResource.class);
    //    }
    if (resource != null && resource.getType() == IResource.FILE) {
      IFile iFile = (IFile) resource;
      file = new File(iFile.getLocation().toString());
    }

    if (file != null && file.getName().toLowerCase().endsWith(".java")) {
      list.add(new ReadableDocumentFileAdapter(file));
    }
  }

  protected final ISelection getSelection() {
    ISelectionProvider selectionProvider = getWorkbenchPart().getSite().getSelectionProvider();
    return selectionProvider.getSelection();
  }

  //  protected File getJarFile(IAdaptable adaptable) {
  //    JarPackageFragmentRoot jpfr = (JarPackageFragmentRoot) adaptable;
  //    File selected =  (File)  jpfr.getPath().makeAbsolute().toFile();
  //    if ( !((File)selected).exists() ) {
  //      File projectFile = new File(jpfr.getJavaProject().getProject().getLocation().toOSString());
  //      selected = new File( projectFile.getParent() + selected.toString() );
  //    }
  //    return selected;
  //  }

  public void promptForSaveAndConvertDocuments(JavaDocumentList documentList) {
    promptForSaveAndConvertDocuments(documentList, new JavaSource2HTMLConverter());
  }

  public void promptForSaveAndConvertDocuments(JavaDocumentList documentList, IJavaSourceConverter converter) {
    if (documentList == null || documentList.getSize() == 0) {
      EclipseUtilities.out(getShell(), "No source code to convert found.");
      return;
    }

    String[] filterNames = new String[]{ "HTML Files", "All Files" };
    String[] filterExtensions = new String[]{ "*." + converter.getMetaData().getDefaultFileExtension(), "*.*" };
    File file = EclipseUtilities.promptForSaveFile(documentList.getName()
        + "."
        + converter.getMetaData().getDefaultFileExtension(), filterNames, filterExtensions);
    if (file == null) {
      return;
    }

    if (file.exists()) {
      if (!MessageDialog.openConfirm(getShell(), "File exists", file
          + " already exists. Do you want to replace it?")) {
        return;
      }
    }

    //    if (!file.canWrite()){
    //      Shell shell = new Shell();
    //      ErrorDialog.openError(shell, "Output error", "Unable to write to file "+file.getName(), null);
    //    }

    BufferedWriter writer = null;
    try {
      writer = new BufferedWriter(new FileWriter(file));
      convertReadableDocuments(documentList, converter, writer);
      writer.flush();
    }
    catch (IOException e) {
      ErrorDialog.openError(getShell(), "Output error", "Error writing file: " + file, null);
    }
    finally {
      IoUtilities.close(writer);
    }
  }

  public void convertReadableDocuments(
      final JavaDocumentList documents,
      final IJavaSourceConverter converter,
      Writer writer) throws IOException {

    final BufferedWriter bw = new BufferedWriter(writer);

    try {
      IRunnableWithProgress op = new IRunnableWithProgress() {
        public void run(IProgressMonitor progressMonitor) throws InvocationTargetException, InterruptedException {
          progressMonitor.beginTask("Converting files", documents.getSize());
          JavaSourceConversionOptions options = Java2HtmlPlugin.getDefault().getConversionOptions();

          try {
            converter.writeDocumentHeader(bw, options, "");
            Iterator iterator = documents.iterator();
            while (iterator.hasNext()) {
              ReadableDocument document = (ReadableDocument) iterator.next();
              progressMonitor.subTask(document.getName());
              try {
                JavaSourceParser parser = new JavaSourceParser(options);
                JavaSource source = parser.parse(document.getReader());
                source.setFileName(document.getName());
                converter.convert(source, options, bw);
              }
              catch (Exception e) {
                System.err.println("Error converting " + document.getName() + ": " + e);
              }
              progressMonitor.worked(1);
              if (progressMonitor.isCanceled()) {
                return;
              }
            }
            converter.writeDocumentFooter(bw, options);
            bw.flush();
          }
          catch (IOException e) {
            throw new InvocationTargetException(e);
          }
          finally {
            progressMonitor.done();
          }
        }
      };
      new ProgressMonitorDialog(getShell()).run(true, true, op);
    }
    catch (InvocationTargetException e) {
      throw (IOException) e.getCause();
    }
    catch (InterruptedException e) {
      // handle cancelation
    }
  }
}